import { filter, map } from 'rxjs/operators';
import { Subject } from 'rxjs/Subject';
import { log } from './misc/Util';
import { signaling } from './proto';
var PING_INTERVAL = 3000;
/* WebSocket error codes */
var MESSAGE_ERROR_CODE = 4000;
var PING_ERROR_CODE = 4001;
var FIRST_CONNECTION_ERROR_CODE = 4002;
/* Preconstructed messages */
var pingMsg = signaling.Message.encode(signaling.Message.create({ ping: true })).finish();
var pongMsg = signaling.Message.encode(signaling.Message.create({ pong: true })).finish();
export var SignalingState;
(function (SignalingState) {
    SignalingState[SignalingState["CONNECTING"] = 0] = "CONNECTING";
    SignalingState[SignalingState["OPEN"] = 1] = "OPEN";
    SignalingState[SignalingState["CONNECTED_WITH_FIRST_MEMBER"] = 2] = "CONNECTED_WITH_FIRST_MEMBER";
    SignalingState[SignalingState["READY_TO_JOIN_OTHERS"] = 3] = "READY_TO_JOIN_OTHERS";
    SignalingState[SignalingState["CLOSED"] = 4] = "CLOSED";
})(SignalingState || (SignalingState = {}));
/**
 * This class represents a door of the `WebChannel` for the current peer. If the door
 * is open, then clients can join the `WebChannel` through this peer. There are as
 * many doors as peers in the `WebChannel` and each of them can be closed or opened.
 */
var Signaling = (function () {
    function Signaling(wc, url) {
        // public
        this.url = url;
        this.state = SignalingState.CLOSED;
        // private
        this.wc = wc;
        this.stateSubject = new Subject();
        this.channelSubject = new Subject();
        this.rxWs = undefined;
        this.pingInterval = undefined;
        this.pongReceived = false;
    }
    Object.defineProperty(Signaling.prototype, "onState", {
        get: function () {
            return this.stateSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Signaling.prototype, "onChannel", {
        get: function () {
            return this.channelSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Notify Signaling server that you had joined the network and ready
     * to join new peers to the network.
     */
    Signaling.prototype.open = function () {
        if (this.state === SignalingState.CONNECTED_WITH_FIRST_MEMBER) {
            this.rxWs.send({ joined: true });
            this.setState(SignalingState.READY_TO_JOIN_OTHERS);
        }
    };
    Signaling.prototype.join = function (key) {
        var _this = this;
        if (this.state === SignalingState.READY_TO_JOIN_OTHERS) {
            throw new Error('Failed to join via signaling: connection with signaling is already opened');
        }
        if (this.state !== SignalingState.CLOSED) {
            this.close();
        }
        this.setState(SignalingState.CONNECTING);
        this.wc.webSocketBuilder.connect(this.getFullURL(key))
            .then(function (ws) {
            _this.setState(SignalingState.OPEN);
            _this.rxWs = _this.createRxWs(ws);
            _this.startPingInterval();
            _this.rxWs.onMessage.subscribe(function (msg) {
                switch (msg.type) {
                    case 'ping':
                        _this.rxWs.pong();
                        break;
                    case 'pong':
                        _this.pongReceived = true;
                        break;
                    case 'isFirst':
                        if (msg.isFirst) {
                            _this.setState(SignalingState.READY_TO_JOIN_OTHERS);
                        }
                        else {
                            _this.wc.webRTCBuilder.connectOverSignaling({
                                onMessage: _this.rxWs.onMessage.pipe(filter(function (m) { return m.type === 'content'; }), map(function (_a) {
                                    var content = _a.content;
                                    return content;
                                })),
                                send: function (m) { return _this.rxWs.send({ content: m }); },
                            })
                                .then(function (ch) {
                                _this.setState(SignalingState.CONNECTED_WITH_FIRST_MEMBER);
                                ch.markAsIntermediry();
                            })
                                .catch(function (err) {
                                _this.rxWs.close(FIRST_CONNECTION_ERROR_CODE, "Failed to join over Signaling: " + err.message);
                            });
                        }
                        break;
                }
            });
        })
            .catch(function (err) { return _this.setState(SignalingState.CLOSED); });
    };
    /**
     * Close the `WebSocket` with Signaling server.
     */
    Signaling.prototype.close = function () {
        if (this.rxWs) {
            this.rxWs.close(1000);
        }
    };
    Signaling.prototype.setState = function (state) {
        var _this = this;
        if (this.state !== state) {
            this.state = state;
            this.stateSubject.next(state);
            if (state === SignalingState.READY_TO_JOIN_OTHERS) {
                this.wc.webRTCBuilder.onChannelFromSignaling({
                    onMessage: this.rxWs.onMessage.pipe(filter(function (msg) { return msg.type === 'content'; }), map(function (_a) {
                        var content = _a.content;
                        return content;
                    })),
                    send: function (msg) { return _this.rxWs.send({ content: msg }); },
                }).subscribe(function (ch) { return _this.channelSubject.next(ch); });
            }
        }
    };
    Signaling.prototype.startPingInterval = function () {
        var _this = this;
        this.rxWs.ping();
        this.pingInterval = setInterval(function () {
            if (_this.state !== SignalingState.CLOSED) {
                if (!_this.pongReceived) {
                    clearInterval(_this.pingInterval);
                    _this.rxWs.close(PING_ERROR_CODE, 'Signaling is not responding');
                }
                else {
                    _this.pongReceived = false;
                    _this.rxWs.ping();
                }
            }
        }, PING_INTERVAL);
    };
    Signaling.prototype.createRxWs = function (ws) {
        var _this = this;
        var subject = new Subject();
        ws.binaryType = 'arraybuffer';
        ws.onmessage = function (evt) {
            try {
                subject.next(signaling.Message.decode(new Uint8Array(evt.data)));
            }
            catch (err) {
                ws.close(MESSAGE_ERROR_CODE, err.message);
            }
        };
        ws.onerror = function (err) { return subject.error(err); };
        ws.onclose = function (closeEvt) {
            clearInterval(_this.pingInterval);
            _this.setState(SignalingState.CLOSED);
            if (closeEvt.code === 1000) {
                subject.complete();
            }
            log.info("Connection with Signaling '" + _this.url + "' closed: " + closeEvt.code + ": " + closeEvt.reason);
        };
        return {
            onMessage: subject.asObservable(),
            send: function (msg) {
                if (ws.readyState !== WebSocket.CLOSING && ws.readyState !== WebSocket.CLOSED) {
                    ws.send(signaling.Message.encode(signaling.Message.create(msg)).finish());
                }
            },
            ping: function () {
                if (ws.readyState !== WebSocket.CLOSING && ws.readyState !== WebSocket.CLOSED) {
                    ws.send(pingMsg);
                }
            },
            pong: function () {
                if (ws.readyState !== WebSocket.CLOSING && ws.readyState !== WebSocket.CLOSED) {
                    ws.send(pongMsg);
                }
            },
            close: function (code, reason) {
                if (code === void 0) { code = 1000; }
                if (reason === void 0) { reason = ''; }
                ws.onclose = undefined;
                ws.close(code, reason);
                _this.setState(SignalingState.CLOSED);
                clearInterval(_this.pingInterval);
                subject.complete();
            },
        };
    };
    Signaling.prototype.getFullURL = function (params) {
        if (this.url.endsWith('/')) {
            return this.url + params;
        }
        else {
            return this.url + '/' + params;
        }
    };
    return Signaling;
}());
export { Signaling };
